<?php

declare(strict_types=1);

namespace Drupal\mcp\Controller;

use Drupal\Core\Utility\Error;
use Drupal\jsonrpc\Exception\JsonRpcException;
use Drupal\mcp\Config\McpSettings;
use Drupal\mcp\ResponseHandler\ResponseHandlerFactory;
use Drupal\mcp\Sse\SseStreamHandler;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\jsonrpc\Controller\HttpController;
use Symfony\Component\HttpFoundation\Request;
use Drupal\jsonrpc\Object\Error as JsonRpcError;

/**
 * Returns responses for Model Context Protocol routes.
 */
class McpController extends HttpController {

  /**
   * The SSE stream handler.
   */
  protected SseStreamHandler $sseStreamHandler;

  /**
   * The response handler factory.
   */
  protected ResponseHandlerFactory $responseHandlerFactory;

  /**
   * The MCP settings.
   */
  protected McpSettings $mcpSettings;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    $instance = parent::create($container);

    $instance->sseStreamHandler = $container->get('mcp.sse.stream_handler');
    $instance->responseHandlerFactory = $container->get(
      'mcp.response_handler.factory'
    );
    $instance->mcpSettings = $container->get('mcp.settings');
    $instance->handler = $container->get('mcp.jsonrpc.handler');

    return $instance;
  }

  /**
   * SSE endpoint.
   */
  public function get() {
    try {
      if (!$this->mcpSettings->isSseEnabled()) {
        throw new \Exception('SSE is not enabled');
      }

      return $this->sseStreamHandler->createStream();
    }
    catch (\Exception $e) {
      Error::logException(
        $this->getLogger('mcp'),
        $e,
        'An error occurred while creating the SSE stream: @message',
        ['@message' => $e->getMessage()]
      );

      return $this->exceptionResponse(
        JsonRpcException::fromError(
          new JsonRpcError(
            JsonRpcError::INTERNAL_ERROR,
            'SSE is not enabled'
          )
        )
      );
    }
  }

  /**
   * Post endpoint for receiving messages from the client.
   */
  public function post(Request $request) {
    $rpcResponse = $this->resolve($request);
    $clientId = $request->query->get('sessionId');

    try {
      return $this->responseHandlerFactory
        ->getHandler()
        ->handleResponse($rpcResponse, $clientId);
    }
    catch (\Exception $e) {
      Error::logException(
        $this->getLogger('mcp'),
        $e,
        'An error occurred while handling the SSE response for client %client_id: @message',
        [
          '%client_id' => $clientId,
          '@message'   => $e->getMessage(),
        ]
      );

      return $this->exceptionResponse(
        JsonRpcException::fromPrevious($e)
      );
    }
  }

}
