<?php

namespace Drupal\mcp\Sse;

use Drupal\Component\Uuid\UuidInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Url;
use Drupal\Core\Utility\Error;
use Drupal\mcp\Storage\ClientMessageStorageInterface;
use Symfony\Component\HttpFoundation\StreamedResponse;

/**
 * Handles the SSE stream for the MCP server.
 */
class SseStreamHandler {

  /**
   * Time in seconds before considering a client disconnected.
   */
  const MCP_HEARTBEAT_TIMEOUT = 300;

  /**
   * Interval in seconds between heartbeat messages.
   */
  const MCP_HEARTBEAT_MESSAGE_INTERVAL = 30;

  /**
   * The logger instance.
   */
  protected LoggerChannelInterface $logger;

  public function __construct(
    private readonly ClientMessageStorageInterface $storage,
    private readonly UuidInterface $uuid,
    LoggerChannelFactoryInterface $logger_factory,
  ) {
    $this->logger = $logger_factory->get('mcp');
  }

  /**
   * Creates a new SSE stream.
   *
   * @throws \Exception
   */
  public function createStream(): StreamedResponse {
    $clientId = $this->uuid->generate();

    ignore_user_abort(FALSE);

    $this->storage->connectClient($clientId);

    register_shutdown_function(function () use ($clientId) {
      $this->cleanupAndBreak($clientId);
    });

    $response = new StreamedResponse();
    $response->headers->set('Content-Type', 'text/event-stream');
    $response->headers->set('Cache-Control', 'no-cache');
    $response->headers->set('Connection', 'keep-alive');
    $response->headers->set('X-Accel-Buffering', 'no');

    $response->setCallback(function () use ($clientId) {
      while (ob_get_level() > 0) {
        ob_end_clean();
      }
      ob_implicit_flush();

      try {
        $this->sendSseMessage(
          'endpoint',
          Url::fromRoute('mcp.post', [], [
            'query' => ['sessionId' => $clientId],
          ])->toString()
        );

        $lastHeartbeat = $lastHeartbeatMessage = time();

        while (!connection_aborted()) {
          if ((time() - $lastHeartbeat) >= self::MCP_HEARTBEAT_TIMEOUT
            || !$this->storage->isClientConnected($clientId)
          ) {
            $this->logger->info(
              'Client %client_id connection ended: @reason',
              [
                '%client_id' => $clientId,
                '@reason'    => connection_aborted()
                  ? 'connection_aborted'
                  : ((time() - $lastHeartbeat) >= self::MCP_HEARTBEAT_TIMEOUT
                    ? 'timeout' : 'disconnected'),
              ]
            );
            break;
          }

          if ((time() - $lastHeartbeatMessage)
            >= self::MCP_HEARTBEAT_MESSAGE_INTERVAL
          ) {
            $this->sendSseMessage(
              'heartbeat', json_encode(['timestamp' => time()])
            );
            $lastHeartbeatMessage = time();
            continue;
          }

          if ($message = $this->storage->getMessage($clientId)) {
            $lastHeartbeat = time();
            $this->sendSseMessage('message', $message);
            $this->storage->deleteMessage($clientId);
          }

          usleep(100000);
        }
      }
      catch (\Exception $e) {
        Error::logException(
          $this->logger,
          $e,
          'SSE stream error for client %client_id: @message',
          [
            '%client_id' => $clientId,
            '@message'   => $e->getMessage(),
          ]
        );
      }
      finally {
        $this->cleanupAndBreak($clientId);
      }
    });

    return $response;
  }

  /**
   * Sends an SSE message.
   *
   * @throws \Exception
   */
  protected function sendSseMessage(string $event, string $data): void {
    if (connection_aborted()) {
      throw new \Exception('Connection aborted while sending message');
    }

    echo "event: $event\n";
    echo "data: $data\n\n";
  }

  /**
   * Cleanup client resources and break the connection.
   *
   * @param string $clientId
   *   The client ID.
   *
   * @throws \Exception
   */
  protected function cleanupAndBreak(string $clientId): void {
    if ($this->storage->isClientConnected($clientId)) {
      $this->logger->info(
        'Cleaning up client %client_id connection',
        [
          '%client_id' => $clientId,
        ]
      );

      $this->storage->disconnectClient($clientId);
      $this->storage->deleteMessage($clientId);
    }
  }

}
