<?php

namespace Drupal\mcp\Storage;

use Drupal\Core\Database\Connection;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Utility\Error;

/**
 * Database storage for client messages.
 */
class DatabaseClientMessageStorage implements ClientMessageStorageInterface {

  // Message time-to-live in seconds.
  const MESSAGE_TTL = 300;

  /**
   * The logger instance.
   */
  protected LoggerChannelInterface $logger;

  /**
   * Constructor.
   */
  public function __construct(
    private readonly Connection $database,
    LoggerChannelFactoryInterface $logger_factory,
  ) {
    $this->logger = $logger_factory->get('mcp');
  }

  /**
   * {@inheritdoc}
   */
  public function isClientConnected(string $clientId): bool {
    $exists = $this->database->select('mcp_messages', 'm')
      ->fields('m', ['client_id'])
      ->condition('client_id', $clientId)
      ->condition('created', time() - self::MESSAGE_TTL, '>')
      ->countQuery()
      ->execute()
      ->fetchField();

    return (bool) $exists;
  }

  /**
   * {@inheritdoc}
   */
  public function connectClient(string $clientId): void {
    $this->database->merge('mcp_messages')
      ->key('client_id', $clientId)
      ->fields([
        'message' => '{}',
        'created' => time(),
      ])
      ->execute();
  }

  /**
   * {@inheritdoc}
   */
  public function disconnectClient(string $clientId): void {
    $this->database->delete('mcp_messages')
      ->condition('client_id', $clientId)
      ->execute();
  }

  /**
   * {@inheritdoc}
   */
  public function storeMessage(string $clientId, string $message): void {
    $this->database->merge('mcp_messages')
      ->key('client_id', $clientId)
      ->fields([
        'message' => json_encode($message, JSON_THROW_ON_ERROR),
        'created' => time(),
      ])
      ->execute();
  }

  /**
   * {@inheritdoc}
   */
  public function getMessage(string $clientId): mixed {
    $record = $this->database->select('mcp_messages', 'm')
      ->fields('m', ['message'])
      ->condition('client_id', $clientId)
      ->condition('created', time() - self::MESSAGE_TTL, '>')
      ->execute()
      ->fetchField();

    return $record ? json_decode($record, TRUE, 512, JSON_THROW_ON_ERROR)
      : NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function deleteMessage(string $clientId): void {
    $this->database->update('mcp_messages')
      ->fields([
        'message' => '{}',
      ])
      ->condition('client_id', $clientId)
      ->execute();
  }

  /**
   * Cleanup old messages.
   */
  public function cleanup(): void {
    try {
      $this->database->delete('mcp_messages')
        ->condition('created', time() - self::MESSAGE_TTL, '<')
        ->execute();
    }
    catch (\Exception $e) {
      Error::logException(
        $this->logger,
        $e,
        'Failed to cleanup old messages: @message',
        [
          '@message' => $e->getMessage(),
        ]
      );
    }
  }

}
